﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Diagnostics;
using RootPro.RootProCAD;
using RootPro.RootProCAD.Command;
using RootPro.RootProCAD.Geometry;
using RootPro.RootProCAD.UI;

//////////////////////////////////////////////////////////////////
// 分割割付のアドイン
// 2 つの図形の間に、等間隔に図形を作成します。
//////////////////////////////////////////////////////////////////

namespace RCAddInAllocationCS
{
    [System.AddIn.AddIn("AppAddIn", Version = "1.0", Publisher = "", Description = "")]
    public partial class AppAddIn
    {
        string AllocationCommandName = "RootPro.AddIn.AllocationCommand";     //コマンド名 
        
        private void AppAddIn_Startup(object sender, EventArgs e)
        {
            CommandManager.AddMacroCommand("RCAddInAllocationCS", "分割割付", CreateAllocationShape);

            //結合コマンドを追加する
            Command command = CommandManager.AddCommand(AllocationCommandName, "分割割付");

            //パラメータを追加する
            for (int i = 0; i < 2; i++)
            {
                CommandParameterGroup commandParameterGroup = command.ParameterGroups.Add(ParameterItemType.Shape, i, "図形 " + (i + 1));
                ShapeParameterItem parameterItem = (ShapeParameterItem)commandParameterGroup.CurrentItem; // パラメータアイテム
                parameterItem.ShapeTypeFilter = ShapeTypeFilters.Line | ShapeTypeFilters.Circle;      //線分と連続線のみ
                parameterItem.Description = "線分、又は円を指定します。同じ図形タイプを選択する必要があります。";
            }

            command.ParameterGroups.Add(ParameterItemType.Integer, 2, "分割数");
            IntegerParameterItem integerParameterItem = (IntegerParameterItem)command.GetParameterItemByID(2);
            integerParameterItem.Description = "分割数を設定します。（ 2 ～ 100 )";
            integerParameterItem.Value = 2;
            integerParameterItem.ValueSaveMode = ParameterValueSaveMode.Always;

            //開始イベントと終了イベントの追加
            command.Begin += new CommandBeginEventHandler(command_Begin);
            command.End += new EventHandler(command_End);
        }

        private void AppAddIn_Shutdown(object sender, EventArgs e)
        {
            CommandManager.RemoveMacroCommand(CreateAllocationShape);
            CommandManager.RemoveCommand(AllocationCommandName);
        }

        #region VSTA generated code
        private void InternalStartup()
        {
            this.Startup += new System.EventHandler(AppAddIn_Startup);
            this.Shutdown += new System.EventHandler(AppAddIn_Shutdown);
        }
        #endregion

        /// <summary>
        /// 分割割付コマンド処理
        /// </summary>
        private void CreateAllocationShape()
        {
            //コマンドの実行
            CommandManager.ExecuteCommand(AllocationCommandName);
        }

        /// <summary>
        /// コマンドの開始時に呼ばれる
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void command_Begin(Object sender, CommandBeginEventArgs e)
        {
            Command command = (Command)sender;

            //パラメータ変更イベントの追加
            command.ParameterChanged += new CommandParameterChangedEventHandler(command_ParameterChanged);
        }

        /// <summary>
        /// コマンドの終了時に呼ばれる
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void command_End(Object sender, EventArgs e)
        {
            Command command = (Command)sender;

            //パラメータ変更イベントの削除
            command.ParameterChanged -= new CommandParameterChangedEventHandler(command_ParameterChanged);
        }

        /// <summary>
        /// 分割割付を実行する
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void command_ParameterChanged(Object sender, CommandParameterChangedEventArgs e)
        {
            Document doc = ActiveDocument;         //ドキュメントの取得

            Command command = (Command)sender;
            int id = e.ParameterID;// 変更のあったパラメータID
            ParameterItem parameterItem = command.GetParameterItemByID(id); // パラメータアイテム

            //図形１の取得
            ShapeParameterItem shapeParameterItem1 = (ShapeParameterItem)command.GetParameterItemByID(0);
            ShapeParameterItem shapeParameterItem2 = (ShapeParameterItem)command.GetParameterItemByID(1);
            if (shapeParameterItem1.IsEmpty)
            {
                if (!shapeParameterItem2.IsEmpty)
                {
                    Shape shape2a = shapeParameterItem2.Shape.Shape;
                    if (shape2a is LineShape)
                        shapeParameterItem2.ShapeTypeFilter = ShapeTypeFilters.Line;
                    else
                        shapeParameterItem2.ShapeTypeFilter = ShapeTypeFilters.Circle;
                }
                else
                {
                    shapeParameterItem1.ShapeTypeFilter = ShapeTypeFilters.Circle | ShapeTypeFilters.Line;
                    shapeParameterItem2.ShapeTypeFilter = ShapeTypeFilters.Circle | ShapeTypeFilters.Line;
                }
                return;
            }
            Shape shape1 = shapeParameterItem1.Shape.Shape;

            //図形２の取得
            if (shapeParameterItem2.IsEmpty)
            {
                if (shape1 is LineShape)
                    shapeParameterItem2.ShapeTypeFilter = ShapeTypeFilters.Line;
                else
                    shapeParameterItem2.ShapeTypeFilter = ShapeTypeFilters.Circle;
                return;
            }
            Shape shape2 = shapeParameterItem2.Shape.Shape;

            //図形が同じタイプか？
            if (shape1.GetType() != shape2.GetType())
            {
                MessageBox.Show("同じタイプの図形を選択してください");
                return;
            }

            //分割数の取得
            IntegerParameterItem integerParameterItem = (IntegerParameterItem)command.GetParameterItemByID(2);
            if (integerParameterItem.IsEmpty)
                return;
            int dividCount = integerParameterItem.Value;
            if (dividCount < 2 || dividCount > 100)
            {
                MessageBox.Show("分割数は 2 ～ 100 の範囲を指定してください。");
                return;
            }


            // UNDO 可能な操作を開始
            doc.UndoManager.BeginUndoUnit();

            //時間計測
            Stopwatch sw = new Stopwatch();
            sw.Start();

            //分割図形の作成
            CreateAllocationShape(shape1, shape2, dividCount);

            //図形パラメータを空にする
            command.ParameterChanged -= new CommandParameterChangedEventHandler(command_ParameterChanged);
            shapeParameterItem1.SetEmpty();
            shapeParameterItem2.SetEmpty();
            command.ParameterChanged += new CommandParameterChangedEventHandler(command_ParameterChanged);

            sw.Stop();

            // UNDO 可能な操作を終了
            doc.UndoManager.EndUndoUnit();


            long millisec = sw.ElapsedMilliseconds;
            Trace.WriteLine("処理時間 " + millisec + " (msec)");
            Trace.Flush();

        }

        /// <summary>
        /// 分割割付の実行
        /// </summary>
        /// <param name="shape1"></param>
        /// <param name="shape2"></param>
        /// <param name="dividCount"></param>
        private void CreateAllocationShape(Shape shape1, Shape shape2, int dividCount)
        {
            Document doc = ActiveDocument;         //ドキュメントの取得
            Drawing drawing = doc.CurrentDrawing;  //部分図の取得


            Type type = shape1.GetType();
            if (type.Name == "CircleShapeClass")        //円の分割割付
            {
                CircleShape circle1 = (CircleShape)shape1;
                CircleShape circle2 = (CircleShape)shape2;

                for (int i = 0; i < dividCount - 1; i++)
                {
                    double cx = circle1.Center.X + (circle2.Center.X - circle1.Center.X) * (i + 1) / dividCount;
                    double cy = circle1.Center.Y + (circle2.Center.Y - circle1.Center.Y) * (i + 1) / dividCount;
                    double radius = circle1.Radius + (circle2.Radius - circle1.Radius) * (i + 1) / dividCount;

                    Point2d centerPoint = Geometry.CreatePoint(cx, cy);

                    CircleShape circleShape = drawing.Shapes.AddCircle(centerPoint, radius);
                }

            }
            else if (type.Name == "LineShapeClass")     //線分の分割割付
            {
                LineShape line1 = (LineShape)shape1;
                LineShape line2 = (LineShape)shape2;


                double sx1 = line1.StartPoint.X;
                double sy1 = line1.StartPoint.Y;
                double ex1 = line1.EndPoint.X;
                double ey1 = line1.EndPoint.Y;

                double sx2 = line2.StartPoint.X;
                double sy2 = line2.StartPoint.Y;
                double ex2 = line2.EndPoint.X;
                double ey2 = line2.EndPoint.Y;

                double angle1 = GetAngle(sx1, sy1, ex1, ey1);
                double angle2 = GetAngle(sx2, sy2, ex2, ey2);


                Trace.WriteLine((angle1 * 180.0 / Math.PI).ToString("0.0") + "-" + (angle2 * 180.0 / Math.PI).ToString("0.0"));
                Trace.Flush();

                if (Math.Abs(angle2 - angle1) > Math.PI / 2 && Math.Abs(angle2 - angle1) < Math.PI * 3 / 2)
                {
                    //角度が開いているので、逆点で対応する
                    sx2 = line2.EndPoint.X;
                    sy2 = line2.EndPoint.Y;
                    ex2 = line2.StartPoint.X;
                    ey2 = line2.StartPoint.Y;
                }

                for (int i = 0; i < dividCount - 1; i++)
                {
                    double x1 = sx1 + (sx2 - sx1) * (i + 1) / dividCount;
                    double y1 = sy1 + (sy2 - sy1) * (i + 1) / dividCount;
                    double x2 = ex1 + (ex2 - ex1) * (i + 1) / dividCount;
                    double y2 = ey1 + (ey2 - ey1) * (i + 1) / dividCount;

                    Point2d point1 = Geometry.CreatePoint(x1, y1);
                    Point2d point2 = Geometry.CreatePoint(x2, y2);

                    LineShape lineShape = drawing.Shapes.AddLine(point1, point2);
                }

            }

        }

        /// <summary>
        /// 2点から角度を求める（RAD)
        /// </summary>
        /// <param name="point1"></param>
        /// <param name="point2"></param>
        /// <returns></returns>
        private double GetAngle(double x1, double y1, double x2, double y2)
        {
            double dx = (x2 - x1);
            double dy = (y2 - y1);

            double angle = Math.Atan2(dy, dx);
            //位置により角度を調整する
            if (dx >= 0.0)
            {
                if (dy >= 0.0)
                {
                    if (angle > Math.PI / 2)
                        angle -= Math.PI;
                    else if (angle < 0.0)
                        angle += Math.PI;
                }
                else
                {
                    if (angle > Math.PI / 2)
                        angle -= Math.PI;
                }
            }
            else
            {
                if (dy >= 0.0)
                {
                    if (angle <= Math.PI / 2)
                        angle += Math.PI;
                }
                else
                {
                    if (angle > 0.0)
                        angle -= Math.PI;
                }
            }

            return angle;
        }
    }
}
